# Copyright (C) 2022, Tristan Dagobert <tristan.dagobert@ens-paris-saclay.fr>

# -*- coding: utf-8 -*-
import os
import argparse
import zipfile
import numpy as np
import glob
from os.path import exists, join, realpath, dirname, basename, splitext, isfile
import iio
import cv2
import matplotlib.pyplot as plt
from matplotlib import cm
from scipy.ndimage import sobel
from scipy.stats import chi2
from scipy.linalg import inv
from scipy.sparse.linalg import cg
from scipy.sparse import eye

from sklearn.metrics import jaccard_score, accuracy_score, balanced_accuracy_score
import gmm
import jax
#import numba

key = jax.random.PRNGKey(0)

BORDS = 6

def ma_palette(ndates, nclasses, coords_μ):
    palette = np.zeros((ndates, nclasses, 3))
    mini = np.min(coords_μ)
    maxi = np.max(coords_μ)
        
    for t in np.arange(ndates):
        for c in np.arange(3):
            val = (coords_μ[t, :, c] - mini) / (maxi - mini)
            val = 255 * val 
            palette[t, :, c] = np.array(val, dtype=np.int)
    return palette


def mon_3d(ident, ndates, nclasses, coord_μ):

    palette = ma_palette(ndates, nclasses, coord_μ)

    # affichage 3D des moyennes
    plt.figure(ident)
    plt.clf()
    ax = plt.figure(ident).add_subplot(projection='3d')
    
    ax.set_title('Répartition des centroïdes')
    ax.set_xlabel('R')
    ax.set_ylabel('V')
    ax.set_zlabel('B')
    lim = 5
    ax.set_xlim(-lim, lim)
    ax.set_ylim(-lim, lim)
    ax.set_zlim(-lim, lim)
    for t in np.arange(ndates):
        for classe in np.arange(nclasses):
            x = coord_μ[t, classe, 0]
            y = coord_μ[t, classe, 1]
            z = coord_μ[t, classe, 2]
            print(x, y, z)
            couleur = palette[t, classe, :] / 255.0
            couleur = np.expand_dims(couleur, axis=0)
#            print(type(couleur), couleur.shape, couleur)
            ax.scatter(x, y, z, zdir='y', c=couleur, label='μ')
    # Customize the view angle so it's easier to see that the scatter points lie
    # on the plane y=0
    ax.legend()
    ax.view_init(elev=20., azim=-35)
    plt.draw()

    return


def projeter_centres(X, serie_mat_μ, outdir, bandes):
    """
    X : np.array
        Échantillon ndim=(nsample, nbandes)
    serie_mat_μ : list(np.array)
        Liste des centres, ndim=(nclasses, nbandes)[ndates]
    outdir : str
        Répertoire de sortie
    bandes : list
        Liste des bandeshttps://github.com/centreborelli/fast-gmm
    """
    
    # analyse en composantes principales
    print("dimension vecteur X", X.T.shape)

    cov = np.cov(X.T)
    print("dimension matrice covariance de (I-μ)", cov.shape)

    # les vecteurs propres (en colonnes) sont triés par importance
    eig_val, eig_vec = np.linalg.eig(cov)

    print("dimension matrice des vecteurs propres E", eig_vec.shape)
    print('Eigenvalues = ', eig_val)
#    vec = 10 * np.random.rand(13)
#    print("aléatoire", vec)
    # réorientation scolaire des vecteurs propres dans un repère direct
    _,  nclasses  = eig_vec.shape


    for n in np.arange(3):
        if np.sum(eig_vec[:, n]) > 0:
            eig_vec[:, n] = eig_vec[:, n]
        else:
            print("None inversion du PCA %d" % (n+1))
            eig_vec[:, n] = -1.0 * eig_vec[:, n]
                                    
    plt.figure(1)
    plt.clf()
    plt.title("Coordonnées des vecteurs propres")
    plt.ylim([-2, 2])
    plt.grid()
    plt.xticks(np.arange(len(bandes)), labels=bandes)
    plt.plot(eig_vec[:, 0].T, label="PCA 1", color="r")
    plt.plot(eig_vec[:, 1].T, label="PCA 2", color="g")
    plt.plot(eig_vec[:, 2].T, label="PCA 3", color="b")
    plt.legend()
    plt.savefig(join(outdir, "pca_%d.png" % (0)))
        
    # projection des centres des gaussiennes sur les composantes principales
    ndates = len(serie_mat_μ)
    nclasses, nbandes = serie_mat_μ[0].shape
    
    serie_coord_μ = np.zeros((ndates, nclasses, 3))
    for t, mat_μ in enumerate(serie_mat_μ):
#        print("=== date %d" % (t))
#        print("mat_μ")        
#        print(mat_μ)
#        print("eig_vec")
#        print(eig_vec)
        πμ = mat_μ @ eig_vec  # πμ=(nclasses, nbandes)
        serie_coord_μ[t] = πμ[:, 0:3]
#        print("serie_coord_μ[t]")
#        print(serie_coord_μ[t])
        
    return serie_coord_μ


def get_standardized_data(series, masque, stand, gt_series=None):
    """
    Parameters
    ----------
    series : dict(list(str)), ndim=(nbandes, ndates)
        Listes des fichiers.
    masque : np.array(bool), ndim=(nlig, ncol)
        Masque des zones interdites.
    stand : str
        Kind of standardization, global or date by date.
    gt_series: list(str), ndim=(ndates)
        Liste des fichiers des masques annotés.

    Return
    ------
    X : np.array, ndim=(ndates, ≤nlig*ncol, nbandes)
        Matrix representing the feature sample at different dates.
    """

    dates = list(series.values())[0]
    nbandes = len(series)

    if masque is None:
        for s in series:
            nlig, ncol, ncan = iio.read(series[s][0]).shape
            msq = np.ones(nlig* ncol, dtype=bool)
            break
    else:
        (nlig, ncol) = masque.shape
        msq = masque.reshape((nlig * ncol))
        
    X = None
    for n in np.arange(len(dates)):

        x = None
        for s in series:
            fichier = series[s][n]
            print(fichier)
            im = iio.read(fichier)

            # the image contains a single band
            (nlig, ncol, ncan) = im.shape

            # transform as a colum vector
            im = im.reshape((nlig * ncol))

            # drop the useless pixels
            im = im[msq]
            im = np.expand_dims(im, axis=1)

            # feature concatenation ndim=(nsample, nbandes)
            if x is None:
                x = np.copy(im)
            else:
                x = np.concatenate((x, im), axis=1)

        if stand == 'D':
            # standardisation locale
            nbandes = x.shape[1]
            for n in np.arange(nbandes):
                xx = x[:, n]
                μ, σ = np.mean(xx), np.std(xx)
                x[:, n] = (xx - μ) / σ
        
        # date concatenation ndim=(ndates, nsample, nbandes)
        if X is None:
            X = np.expand_dims(x, axis=0)
        else:
            X = np.concatenate((X, np.expand_dims(x, axis=0)), axis=0)

    if stand == 'G':
        # standardisation globale
        nbandes = X.shape[2]
        for n in np.arange(nbandes):
            x = X[:, :, n]
            μ, σ = np.mean(x), np.std(x)
            X[:, :, n] = (x - μ) / σ

    if gt_series is None:
        return X

    # annotated masks processing, if any    
    GT = None
    for n in np.arange(len(dates)):
        print(gt_series[n])
        
        gtim = iio.read(gt_series[n])
        print(gtim.shape)
        gtim = gtim[:, :, 0]
        gtim = gtim[BORDS:-BORDS, BORDS:-BORDS]
        nlig, ncol = gtim.shape
        gtim = gtim.reshape((nlig * ncol))

        # drop the useless pixels
        gtim = gtim[msq]
        gtim = gtim.astype(bool)

        if GT is None:
            GT = np.expand_dims(gtim, axis=0)
        else:
            GT = np.concatenate((GT, np.expand_dims(gtim, axis=0)), axis=0)
            
    return X, GT, nlig, ncol

##@numba.jit(nopython=True, parallel=True)
def compute_test(Xn, μ, invΣ, icdfχ2):

    nsample, nbandes = Xn.shape
    M = np.zeros(nsample, dtype=bool)
    
    for i in np.arange(nsample):

        val = (Xn[i]-μ) @ (invΣ @ (Xn[i]-μ).T)
        M[i] = (val <= icdfχ2)
    
    return M


def compute_pλ_at_n0(S, neosurf, nmin, n0, nmax):
    """
    Parameters
    ----------

    Return
    ------
    p_sample: np.array, ndim=(nsample, nbands)
        Spectrum panel sample.
    λ_sample: np.array, ndim=(nsample)
        The λ parameter sample.
    """

    p_sample, λ_sample = None, None
    for ni in np.arange(nmin, n0):
        for nj in np.arange(n0, nmax):

            print(n0, ni, nj)
            xnm1 = S[ni][neosurf]
            xn = S[nj][neosurf]

            xnm1 = xnm1.T
            xn = xn.T         # ndim(xn) = (nbands, nsamples)

            p, λ = calculer_pλ(xn, xnm1)
            if p is None:
                continue

            if p_sample is None:
                p_sample = np.expand_dims(p, axis=0)
            else:
                p_sample = np.concatenate((p_sample,
                                           np.expand_dims(p, axis=0)),
                                          axis=0)
            if λ_sample is None:
                λ_sample = [λ]
            else:
                λ_sample += [λ]
#    p_series += [p_sample]
#    λ_series += [λ_sample]

    return p_sample, λ_sample


def calculer_pλ(xn, xnm1):
    """
    Parameters
    ----------
    xn : np.array, ndim=(nbands, nsamples)
    Return
    ------

    """

    (nbands, m) = xn.shape
    
    β = np.trace(xnm1.T @ xnm1)
    α = np.sum(xnm1, axis=1)
    A = m * np.identity(nbands+1)
    A[0, 0] = β
    A[0, 1:] = α
    A[1:, 0] = α

    γ = np.trace(xn.T @ xnm1)
    s = np.sum(xn, axis=1)
    B = np.zeros(nbands+1)
    B[0] = γ
    B[1:] = s
    
    u, error = cg(A, B, np.zeros(nbands+1))

    if error == 0:
        λ, p = u[0], u[1:]
        return p, λ
    else:
        return None, None
    return


def semiunmixing(outdir, series, gt_series, n0, stand, side):
    """
    Parameters
    ----------
    outdir: str
        Directory where the label images are recorded.
    series: dict(list(str)), ndim=(nbandes, ndates)
        Listes des fichiers.
    gt_series: list(str), ndim=(ndates)
        Liste des fichiers des masques annotés.
    n0 : int
        Time at which one computes the panel spectrum.
    stand: str
        Kind of standardization, global or date by date.
    side : int
        Side of the square patch.

    Return
    ------
    """

    if not exists(outdir):
        os.mkdir(outdir)

    # retrieve the data
    print("retrieve the data")
    S, GT, nlig, ncol = get_standardized_data(series, None, stand,
                                              gt_series=gt_series)
    ndates, nsample, nbandes = S.shape

    # estimation of the panel spectrum according to the ground truth
    zemask = (1 * GT[n0]) - (1 * GT[n0-1])
    neosurf = np.array(zemask, dtype=bool)
    if np.sum(1 * neosurf) < 1:
        print("[FATAL] No new mask at time %d" % (n0))
        exit()
    p_sample, λ_sample = compute_pλ_at_n0(S, neosurf, 0, n0, n0+1)
    p = np.median(p_sample, axis=0)
    x = np.arange(p.shape[0])

    bandes = list(series)
    plt.figure(1)
    plt.title (u"Spectrum $p$ used to identify the patches")
    plt.ylim([-3, 3])
    plt.grid(True)
    plt.scatter(x, p)
    plt.xticks(np.arange(len(bandes)), labels=bandes)
    plt.legend()
    plt.savefig(join(outdir, "spectre.png"))
    
    
    # computation of the λ map series
    λ_files = compute_λ_map_series(S, GT, p, side, nlig, ncol, outdir)

    # record the λ files as rgb files with an appropriate dynamics
        # recherche du maximum de la série
    mini, maxi = -2, +2
    for i, f in enumerate(λ_files):
        im = iio.read(f)
        a = 1.0 * (im - mini) / (maxi - mini)
        a = a.squeeze()
        b = np.uint8(255.0 * cm.jet(a))
        b = b[:, :, 0:3]
        iio.write(join(outdir, "mappa_%d.png" % (i)), b)

    return


def unmixing(outdir, masque, series, gt_series, stand, ndim):
    """
    Parameters
    ----------
    outdir: str
        Directory where the label images are recorded.
    series: dict(list(str)), ndim=(nbandes, ndates)
        Listes des fichiers.
    gt_series: list(str), ndim=(ndates)
        Liste des fichiers des masques annotés.
    stand: str
        Kind of standardization, global or date by date.

    Return
    ------
    """
    if not exists(outdir):
        os.mkdir(outdir)

    # retrieve the data
    print("retrieve the data")
    S, GT = get_standardized_data(series, masque, stand, gt_series=gt_series)
    ndates, nsample, nbandes = S.shape

    p_series, λ_series = {}, {}
    for s in ['full', 'last', 'prev']:
        p_series[s], λ_series[s] = [], []
        
    # différence dt=1 
    for n0 in np.arange(ndates):
        if n0 == 0:
            zemask = np.zeros(ndim)
            iio.write(join(outdir, "neomsq_%d.png" % (n0)), zemask)
            for s in ['full', 'last', 'prev']:
                p_series[s] += [None]
                λ_series[s] += [None]
            continue

        zemask = (1 * GT[n0]) - (1 * GT[n0-1])
        neosurf = np.array(zemask, dtype=bool)
        zemask = 255 * zemask
        zemask = np.reshape(zemask, ndim)
        iio.write(join(outdir, "neomsq_%d.png" % (n0)), zemask)
        # gérer les cas où il n'y a aucune surface nouvelle
        if np.sum(1 * neosurf) < 1:
            for s in ['full', 'last', 'prev']:
                p_series[s] += [None]
                λ_series[s] += [None]
            continue
        
        p_sample, λ_sample = compute_pλ_at_n0(S, neosurf, 0, n0, ndates)
        p_series['full'] += [p_sample]
        λ_series['full'] += [λ_sample]
        
        p_sample, λ_sample = compute_pλ_at_n0(S, neosurf, n0-1, n0, n0+1)
        p_series['last'] += [p_sample]
        λ_series['last'] += [λ_sample]
        
        p_sample, λ_sample = compute_pλ_at_n0(S, neosurf, 0, n0, n0+1)
        p_series['prev'] += [p_sample]
        λ_series['prev'] += [λ_sample]
        
# ALT        p_sample, λ_sample = None, None
# ALT        for ni in np.arange(0, n0):
# ALT            for nj in np.arange(n0, ndates):
# ALT
# ALT                print(n0, ni, nj)
# ALT                Xnm1 = S[ni]
# ALT                Xn = S[nj]
# ALT
# ALT                xnm1 = Xnm1[neosurf]
# ALT                xn = Xn[neosurf]
# ALT
# ALT                xnm1 = xnm1.T
# ALT                xn = xn.T         # ndim(xn) = (nbands, nsamples)
# ALT
# ALT                p, λ = calculer_pλ(xn, xnm1)
# ALT                if p is None:
# ALT                    continue
# ALT                
# ALT                if p_sample is None:
# ALT                    p_sample = np.expand_dims(p, axis=0)
# ALT                else:
# ALT                    p_sample = np.concatenate((p_sample,
# ALT                                               np.expand_dims(p, axis=0)),
# ALT                                              axis=0)
# ALT                if λ_sample is None:
# ALT                    λ_sample = [λ]
# ALT                else:
# ALT                    λ_sample += [λ]
# ALT        p_series += [p_sample]
# ALT        λ_series += [λ_sample]

    # statistics to prepare the box plots
    if stand == "N":
        bornes = [-300, 300]
    else:
        bornes = [-3, 3]
        
    nom_bandes = list(series)
    for s in ['full', 'last', 'prev']:
        for t, p_sample in enumerate(p_series[s]):
            yrange = bornes
            title = "Panel spectrum"
            boxploter(p_sample, yrange, nbandes, nom_bandes, "boxplot", outdir,
                      t, title, s)
    T = t
    for t in np.arange(T):
        plt.figure(t)
        plt.clf()
#        plt.close(t)
        
    for s in ['full', 'last', 'prev']:
        for t, λ_sample in enumerate(λ_series[s]):            
            yrange = [-1.5, 1.5]
            title = "Parameter λ"
            boxploter(λ_sample, yrange, 1, '', "lambplot", outdir, t, title, s)
    return


def boxploter(sample, yrange, nbandes, nom_bandes, pfx, outdir, t, title, strat):

    index = {'full': 3, 'last': 1, 'prev': 2}
    subtitle = {'full': "SPS according to all (ni, nj) s.t. ni < n0, 0 <= nj.",
                'last': "SPS according to pair (n0-1, n0)",
                'prev': "SPS according to pairs (ni, n0) s.t ni < n0"}
            
    if sample is None:
        plt.figure(t)
        plt.title("No sample")
        plt.grid(True)
        plt.ylim(yrange)
        plt.xticks(np.arange(1, nbandes+1), labels=nom_bandes)
        plt.legend()
        plt.savefig(join(outdir, "%s_%d.png" % (pfx, t)))
        return

    nsample = len(sample)
    data = np.zeros((nsample, nbandes))
    for i, p in enumerate(sample):
        data[i,:] = p

        
    plt.figure(t)
#    plt.title(title)
    graphe = plt.subplot(3, 1, index[strat])
    if "λ" in title:
        graphe.set_title(subtitle[strat].replace("SPS", "λ"))
    else:
        graphe.set_title(subtitle[strat])
    graphe.grid(True)
    graphe.set_ylim(yrange)
    graphe.set_xticks(np.arange(1, nbandes+1), labels=nom_bandes)
    graphe.boxplot(data)
    graphe.legend()
    plt.savefig(join(outdir, "%s_%d.png" % (pfx, t)))

    return


def growing_iteratively(outdir, series, masque, gt_series, stand, quantile):
    """
    Parameters
    ----------
    outdir: str
        Directory where the label images are recorded.
    series: dict(list(str)), ndim=(nbandes, ndates)
        Listes des fichiers.
    masque : np.array(bool), ndim=(nlig, ncol)
        Masque des zones interdites.
    stand: 
        Kind of standardization.
    quantile:
        Percentage of pixels used to estimate the ellipsoïd.
    """

    if not exists(outdir):
        os.mkdir(outdir)

    # retrieve the data
    S, GT = get_standardized_data(series, masque, stand, gt_series=gt_series)
    ndates, nsample, nbandes = S.shape
    print("S.shape", S.shape)
    print("GT.shape", GT.shape)
    
    # compute the quantile function
    χ2 = chi2(nbandes) 
    icdfχ2 = χ2.ppf(quantile)
    print(icdfχ2)
    
    nlig, ncol = masque.shape
    M = np.zeros((ndates, nlig, ncol))
    msk = np.zeros((ndates, nlig, ncol))
    
    masque = masque.reshape((nlig*ncol))
    print("ndates, nlig, ncol", ndates, nlig, ncol)
    # compute the Mahalanobis values for each point
    for n in np.arange(ndates):
        Xn = S[n]
        if n > 0:
            gt = V.flatten()
        else:
            gt = GT[n]
        xn = Xn[gt] # ndim(xn) = (nsample, bandes)

        print("Xn.shape", Xn.shape)
        print("xn.shape", xn.shape)
        μ = np.mean(xn, axis=0)
        Σ = np.cov(xn.T)

        print("μ.shape, Σ.shape", μ.shape, Σ.shape)

        invΣ = inv(Σ)

        Mn = compute_test(Xn, μ, invΣ, icdfχ2)
        print("Mn.shape", Mn.shape)
        V = np.zeros((nlig*ncol), dtype=bool)
        print("V.shape", V.shape)
        print("masque.shape", masque.shape)
        V[masque] = Mn
        M[n] = V.reshape((nlig, ncol))
        
        msk[n] = 255 * np.array(M[n], dtype=int)
        nom = join(outdir, "growing%d.png" % (n))
        iio.write(nom, msk[n])
    return M, GT


def growing(outdir, series, masque, gt_series, stand, quantile):
    """
    Parameters
    ----------
    outdir: str
        Directory where the label images are recorded.
    series: dict(list(str)), ndim=(nbandes, ndates)
        Listes des fichiers.
    masque : np.array(bool), ndim=(nlig, ncol)
        Masque des zones interdites.
    stand: 
        Kind of standardization.
    quantile:
        Percentage of pixels used to estimate the ellipsoïd.

    Return
    ------
    M : np.array(bool), ndim(M)=(ndates, nlig, ncol)
        List of Boolean masks.
    GT : np.array(bool), ndim(M)=(ndates, nlig, ncol)
        List of Boolean grund truth masks.
    """

    if not exists(outdir):
        os.mkdir(outdir)

    # retrieve the data
    S, GT = get_standardized_data(series, masque, stand, gt_series=gt_series)
    ndates, nsample, nbandes = S.shape
    print("S.shape", S.shape)
    print("GT.shape", GT.shape)
    
    # compute the quantile function
    χ2 = chi2(nbandes) 
    icdfχ2 = χ2.ppf(quantile)
    print(icdfχ2)
    
    nlig, ncol = masque.shape
    M = np.zeros((ndates, nlig, ncol))
    msk = np.zeros((ndates, nlig, ncol))
    masque = masque.reshape((nlig*ncol))
    
    print("ndates, nlig, ncol", ndates, nlig, ncol)
    # compute the Mahalanobis values for each point
    for n in np.arange(ndates):
        Xn = S[n]
        gt = GT[n]
        xn = Xn[gt] # ndim(xn) = (nsample, bandes)

        print("Xn.shape", Xn.shape)
        print("xn.shape", xn.shape)
        μ = np.mean(xn, axis=0)
        Σ = np.cov(xn.T)

        print("μ.shape, Σ.shape", μ.shape, Σ.shape)

        invΣ = inv(Σ)

        Mn = compute_test(Xn, μ, invΣ, icdfχ2)
        print("Mn.shape", Mn.shape)
        V = np.zeros((nlig*ncol), dtype=bool)
        print("V.shape", V.shape)
        print("masque.shape", masque.shape)
        V[masque] = Mn
        M[n] = V.reshape((nlig, ncol))

        msk[n] = 255 * np.array(M[n], dtype=int)
        nom = join(outdir, "growing%d.png" % (n))
        iio.write(nom, msk[n])
        
    return M, GT
    

def performances(outdir, masques_series, gt_series, quantile):

    ndates, nlig, ncol = masques_series.shape
    
    iou1 = np.zeros(ndates)
    iou2 = np.zeros(ndates)
    iou3 = np.zeros(ndates)
    iou4 = np.zeros(ndates)
    for n, (masque, gt) in enumerate(zip(masques_series, gt_series)):
        print("performances", masque.shape, gt.shape)
        masque = masque.flatten()
#        iou1[n] = jaccard_score(masque.flatten(), gt.flatten(), average='micro')
#        iou2[n] = jaccard_score(masque.flatten(), gt.flatten(), average='macro')
        TP = np.sum(np.logical_and(masque == True, gt == True))
        TN = np.sum(np.logical_and(masque == False, gt == False))
        FP = np.sum(np.logical_and(masque == True, gt == False))
        FN = np.sum(np.logical_and(masque == False, gt == True))

#        a = 255 * masque.reshape((414, 258))
#        b = 255 * gt.reshape((414, 258))

#        iio.write("masque%d.tif" % (n), a)
#        iio.write("gt%d.tif" % (n), b)
        # accuracy
        iou1[n] = 1.0 * (TP + TN) / (TP + FP + TN + FN)

        # precision
        iou2[n] = 1.0 * TP / (TP + FP)

        # recall
        iou3[n] = 1.0 * TP / (TP + FN)
#
#        TPR = 1.0 * TP / (TP + FN)
#        TNR = 1.0 * TN / (TN + FP)
        print(TP, TN, FP, FN)
#        
#        iou4[n] = (TPR + TNR) / 2.0
#        iou3[n] = balanced_accuracy_score(masque.flatten(), gt.flatten())

#        iou4[n] = accuracy_score(masque.flatten(), gt.flatten())
    plt.figure(50)
    plt.clf()
    plt.title("Performances per image with quantile=%5.2f." % (quantile))
    plt.ylim([0, 1.05])
    plt.grid(True)    
 #   plt.plot(iou1, label="micro")
    plt.plot(iou3, label="recall")
    plt.plot(iou1, label="accuracy")
    plt.plot(iou2, label="precision")
    plt.legend()
    plt.savefig(join(outdir, "jaccard.png"))

    return


def classifier_espace(outdir, series, masque, nclasses, methode, stand, quantile):
    """
    Parameters
    ----------
    outdir: str
        Directory where the label images are recorded.
    series: dict(list(str)), ndim=(nbandes, ndates)
        Listes des fichiers.
    masque : np.array(bool), ndim=(nlig, ncol)
        Masque des zones interdites.
    nclasses: int
        Number of classes to segment with.
    method: str
        Kind of segmentation, i.e local or global.
    stand: 
        Kind of standardization.
    quantile:
        Percentage of pixels used to estimate the GMM.

    Return
    ------
    outfiles: list
        …
    bandes: list
        …
    serie_spectres_μ : np.array(ndates, nclasses, nbandes)
        …
    serie_spectres_σ : np.array(ndates, nclasses, nbandes)
        …
    """
    
    if not exists(outdir):
        os.mkdir(outdir)

    # retrieve the data
    S = get_standardized_data(series, masque, stand)
    ndates, nsample, nbandes = S.shape
        
    dates = list(series.values())[0]
    
    # création de l'échantillon
    outfiles = []
                
    series_mat_μ, series_mat_σ, series_plat_σ = [], [], []
    
    serie_spectres_μ = np.zeros((ndates, nclasses, nbandes))
    serie_spectres_σ = np.zeros((ndates, nclasses, nbandes))
    serie_spectres_cσ = np.zeros((ndates, nclasses, nbandes * nbandes))
    serie_coord_gaus = np.zeros((ndates, nclasses, 3))

    # correspondance entre la bande spectrale et son numéro d'index
    bandes = {}
    bandes = list(series)

    (nlig, ncol) = masque.shape
    msq = masque.reshape((nlig*ncol))

    # sélection de 200 000 points (ou ndates x 10 000)
    XXX = np.reshape(S, (ndates * nsample, nbandes))
    ns, _ = XXX.shape
    idx = np.arange(ns)
    np.random.shuffle(idx)

    npts = int(quantile * nlig * ncol * ndates)
    print("subsample of ", npts, "points")
    idx = idx[:npts]  
    XX = XXX[idx, :]

    # apprentissage par rapport à toute la séquence
    if methode == 'G':
        means, covs, weights = gmm.gmm(key, XX, nclasses, n_try=5)
        
    for t in np.arange(ndates):

        if methode == 'D':
            # sélection de 10 000 points
            idx = np.arange(nsample)
            np.random.shuffle(idx)
            npts = int(quantile * nlig * ncol)
            print("subsample of ", npts, "points")
            idx = idx[:npts]
            X = S[t, idx, :]

            # classification selon les valeurs [0, nclasses[
            means, covs, weights = gmm.gmm(key, X, nclasses, n_try=5)

        Y = gmm.predict(S[t], means, covs, weights)
            
        # segmentation classification
        y = -np.ones((nlig*ncol))
        y[msq] = Y
        y = y.reshape((nlig, ncol))

        nom = join(outdir, basename(dates[t]))
        iio.write(nom, y)
        outfiles += [nom]

        # statistiques
        mat_μ = means           # ndim(μ) = (nclasses, nbandes) 
        ten_σ = covs            # ndim(σ) = (nclasses, nbandes, nbandes)            
        plat_σ = ten_σ.reshape((nclasses, nbandes * nbandes))
            
    
        # création des listes
        mat_σ = []
        for c in np.arange(nclasses):
            vec_σ = [ten_σ[c, n, n] for n in np.arange(nbandes)]
            mat_σ += [vec_σ]

        series_mat_μ += [mat_μ]
        series_mat_σ += [mat_σ]
        series_plat_σ += [plat_σ]

    # ACP
    serie_coord_μ = projeter_centres(XX, series_mat_μ, dirname(outdir), bandes)
    mon_3d(0, ndates, nclasses, serie_coord_μ)
    
    #  mise en forme des données de sortie
    for t, (mat_μ, mat_σ, plat_σ) in enumerate(zip(series_mat_μ, series_mat_σ, series_plat_σ)):
        for c in np.arange(nclasses):
            serie_spectres_μ[t, c, :] = mat_μ[c, :]
            serie_spectres_σ[t, c, :] = np.sqrt(mat_σ[c])
            serie_spectres_cσ[t, c, :] = plat_σ[c]
        
    return outfiles, bandes, serie_spectres_μ, serie_spectres_σ, \
        serie_spectres_cσ, serie_coord_μ


def interpoler(series, outdir):
    """
    series : dict(list(str))
        Listes des fichiers, ndim=(nbandes, ndates).
    """
    
    facteurs = {'B01': 6, 'B02': 1, 'B03': 1, 'B04': 1, 'B05': 2, 'B06': 2,
                'B07': 2, 'B08': 1, 'B8A': 2, 'B09': 6, 'B10': 6, 'B11': 2,
                'B12': 2}

    neos = {}

    if not exists(outdir):
        os.mkdir(outdir)

    for s in series:
        serie = series[s]
        fichiers = []
        
        for fichier in serie:
            nom = join(outdir, basename(fichier))
            im = iio.read(fichier).squeeze()

            if facteurs[s] != 1:
                (nlig, ncol) = im.shape
                im = cv2.resize(im, (facteurs[s] * ncol, facteurs[s] * nlig),
#                                interpolation=cv2.INTER_NEAREST)
                                interpolation=cv2.INTER_CUBIC)
            # suppression des bords pour éviter les valeurs nulles dues au
            # zéro-padding de recalage
            im = im[BORDS:-BORDS, BORDS:-BORDS]
            ndim = im.shape
            iio.write(nom, im)
            fichiers += [nom]
                
        neos[s] = fichiers
    return neos, ndim


def calculer_zones_interdites(series, quantile):
    """
    series : dict(list(str))
        Listes des fichiers ndim=(nbandes, ndates)
    quantile : float
        Quantile dans [0; 1]
    msq : np.array(bool) 
        Masque des zones interdites, ndim=(nlig, ncol)
    """
                
    pile = None
    for s in series:
        serie = series[s]
        fichiers = []
        
        for fichier in serie:
            im = iio.read(fichier)

            if quantile == -1.0:
                (nlig, ncol, ncan) = im.shape
                msq = np.ones((nlig, ncol), dtype=bool)
                return msq
            
            if pile is None:
                pile = np.copy(im)
            else:
                pile = np.concatenate((pile, im), axis=2)

    moy = np.mean(pile, axis=2)
    dx = sobel(moy, 0)  # horizontal derivative
    dy = sobel(moy, 1)  # vertical derivative
    mag = np.hypot(dx, dy)
#    iio.write("mag.tif", mag)
    val = np.quantile(mag, 1 - quantile)

    msq = (mag < val)
    return msq


def choisir_palette(ndates, nclasses, coords_μ, dyna):
    """
    coords_μ : np.array(ndates, nclasses, 3)
    """

    palette = np.zeros((ndates, nclasses, 3))
    print("liste des coordonnées μ")
    print(coords_μ)
    if dyna == 'G':
        print("palette c2 bornes absolues")
        mini = np.min(coords_μ)
        maxi = np.max(coords_μ)
    elif dyna == 'I':
        print("palette c2 bornes absolues par canal")
        mini = np.min(coords_μ, axis=(0, 1))
        maxi = np.max(coords_μ, axis=(0, 1))
    print(mini, maxi)
        
    for t in np.arange(ndates):
        if dyna == 'I':
#ALT            mini = np.min(coords_μ[t], axis=0)
#ALT            maxi = np.max(coords_μ[t], axis=0)
#ALT            print(mini.shape, maxi.shape)
#ALT            exit()
            for c in np.arange(3):
                val = (coords_μ[t, :, c] - mini[c]) / (maxi[c] - mini[c])
                val = 255 * val 
                palette[t, :, c] = np.array(val, dtype=int)

        elif dyna == 'G':
#ALT            mini = np.min(coords_μ[t])
#ALT            maxi = np.max(coords_μ[t])

            for c in np.arange(3):
                val = (coords_μ[t, :, c] - mini) / (maxi - mini)
                val = 255 * val 
                palette[t, :, c] = np.array(val, dtype=int)

    return palette


def enregistrer(series, c_rgb, gt, nclasses, outdir, spectres_μ, spectres_σ,
                spectres_cσ, bandes, palette, mat_perm):

    for t, s in enumerate(series):
        im = iio.read(s).squeeze()
        (nlig, ncol) = im.shape
        neo = np.zeros((nlig, ncol, 3))
        if mat_perm is None:
            for c in np.arange(nclasses):
                neo[im == c] = palette[t, c]
        else:
            for c in np.arange(nclasses):
                neo[im == mat_perm[t, c]] = palette[t, c]
                
   
#        nom = join(outdir, basename(s).replace("tiff", "png"))
        nom = join(outdir, "classe_%d.png" % (t))
        print(nom)
        iio.write(nom, neo)
        
    # création des graphes, un par date : #moyennes = ndates
    if isinstance(spectres_μ, np.ndarray):
        ndates, _, _ = spectres_μ.shape
    else:
        ndates = 0
        
    # par classe, spectre du centre de la gaussienne
    for t in np.arange(ndates):
        
        plt.figure(1)
        plt.clf()
        plt.title("Spectre des moyennes")
        plt.ylim([-1.5, 2])
        plt.grid()
        plt.xticks(np.arange(len(bandes)), labels=bandes)

        for c in np.arange(nclasses):
            # couleur de la classe
            couleur = palette[t, c]
            couleur = couleur / 255.0
            # valeurs de moyennes
#            spectre_μ = mat_μ[c, :]
            spectre_μ = spectres_μ[t, c, :]
            # graphe
            plt.plot(spectre_μ, label="Gaus. %d" %(c), color=couleur)
            
        plt.legend()
        plt.savefig(join(outdir, "date_mu_%d.png" % (t)))

    # par classe, spectre de l'écart type de la gaussienne
    for t in np.arange(ndates):

        plt.figure(2)
        plt.clf()
        plt.title("Spectre des écarts types")
        plt.ylim([0, 2.5])
        plt.grid()
        plt.xticks(np.arange(len(bandes)), labels=bandes)

        for c in np.arange(nclasses):
            # couleur de la classe
            couleur = palette[t, c]
            couleur = couleur / 255.0
            # valeurs de variances
         #   spectre_σ = np.sqrt(mat_σ[c])
            spectre_σ = spectres_σ[t, c, :]
            
            # graphe
            plt.plot(spectre_σ, label="Gaus. %d" %(c), color=couleur)

        plt.legend()
        plt.savefig(join(outdir, "date_var_%d.png" % (t)))
            
    # par classe, spectre des écarts types de la gaussienne
    for t in np.arange(ndates):

        plt.figure(3)
        plt.clf()
        plt.title("Spectre de la matrice de covariance (transformée en vecteur)")
        plt.ylim([-2.5, 2.5])
        plt.grid()
#        plt.xticks(np.arange(len(bandes)**2), labels=bandes)

        for c in np.arange(nclasses):
            # couleur de la classe
            couleur = palette[t, c]
            couleur = couleur / 255.0
            # valeurs de variances
         #   spectre_σ = np.sqrt(mat_σ[c])
            spectre_cσ = spectres_cσ[t, c, :]
            
            # graphe
            plt.plot(spectre_cσ, label="Gaus. %d" %(c), color=couleur)

        plt.legend()
        plt.savefig(join(outdir, "date_covar_%d.png" % (t)))

    
    # enregistrement des images RGB
    dates = list(c_rgb.values())[0]
    print(dates)
    
    # création de l'échantillon
    X = []
    for i, _ in enumerate(dates):

        x = None
        for s in ['B04', 'B03', 'B02']:
            fichier = c_rgb[s][i]
            print(fichier)
            im = iio.read(fichier)
            im = im[BORDS:-BORDS, BORDS:-BORDS, :]

            (nlig, ncol, ncan) = im.shape
            
            if x is None:
                x = np.copy(im)
            else:
                x = np.concatenate((x, im), axis=2)
        X += [x]

    for i, im in enumerate(X):
        
        r, g, b = im[:,:,0], im[:,:,1], im[:,:,2]
        r2, g2, b2 = np.copy(r), np.copy(g), np.copy(b)
        r3, g3, b3 = np.sort(r2.flatten()), np.sort(g2.flatten()), np.sort(b2.flatten())
        
        minr, maxr = r3[int(0.001 * r3.size)], r3[int(0.999 * r3.size)]
        ming, maxg = g3[int(0.001 * g3.size)], g3[int(0.999 * g3.size)]
        minb, maxb = b3[int(0.001 * b3.size)], b3[int(0.999 * b3.size)]
        
        r = np.int16(255.0 * (r - minr) / (maxr - minr))
        g = np.int16(255.0 * (g - ming) / (maxg - ming))
        b = np.int16(255.0 * (b - minb) / (maxb - minb))

        r[r<0] = 0
        r[r>255] = 255

        g[g<0] = 0
        g[g>255] = 255

        b[b<0] = 0
        b[b>255] = 255

        r = np.uint8(r)
        g = np.uint8(g)
        b = np.uint8(b)
        
        rgb = np.dstack((r, g, b))
        nom = join(outdir, "rgb_%d.png" % (i))
        iio.write(nom, rgb)

        # incrustation masque
        if gt is None:
            msq = 0.0 * rgb
            nom = join(outdir, "msq_%d.png" % (i))
            iio.write(nom, msq)
        else:
            gtim = iio.read(gt[i])[:, :, 0]
            gtim = gtim[BORDS:-BORDS, BORDS:-BORDS]

            c = np.array([255, 128, 64])
            msq = np.copy(rgb)
            msq[gtim > 0] = c
            nom = join(outdir, "msq_%d.png" % (i))
            iio.write(nom, msq)            
    return


def load_series(cmd):
    """
    Load the spectral filename series selected by the user.

    Parameters
    ----------
    cmd : Namespace
        Contains the parameter of the command line.

    Return
    ------
    series: dict
        Contains the spectral filenames.
    gt_series: list
        Contains the ground truth masks, if any.
    rgb_series: dict
        Contains the spectral filenames.
    """

    # décompactage de l'archive. Lors de ce décompactage, les arborescences
    # sont conservées dans le répertoire de sortie.
    f = zipfile.ZipFile(cmd.zip)
    print("FICHIER :", cmd.zip)
    f.extractall(path=cmd.outdir)
    print("RÉPERTOIRE:", cmd.outdir)
    print("FICHIERS:", f.namelist())
    files = [join(cmd.outdir, fic) for fic in f.namelist()
             if isfile(join(cmd.outdir, fic))]
    myfiles = sorted(files)
    f.close()
    print("[INFO Python] Fichiers myfiles", myfiles)
    print("[INFO Python] Fichier myfiles[0]", os.path.abspath(myfiles[0]))
    
    # convert the Namespace object into a dictionary
    macmd = vars(cmd)

    bandes = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A',
              'B09', 'B10', 'B11', 'B12']

    series = {}
    # search for spectral files
    for b in bandes:
        if eval(macmd[b]) is True:
            series[b] = []
            for x in os.walk(cmd.outdir):
                for y in sorted(glob.glob(join(x[0], '*' + b + '.tif'))):                    
                    series[b] += [y]

    # search for the ground truth masks
    gt_series = None
    for x in os.walk(cmd.outdir):
        if "manual_segmentation_masks" in x[0]:
            gt_series = sorted(glob.glob(join(x[0], '*')))

    # search for the RGB series
    rgb_series = {}
    for b in ['B02', 'B03', 'B04']:
        rgb_series[b] = []
        for x in os.walk(cmd.outdir):
            for y in sorted(glob.glob(join(x[0], '*' + b + '.tif'))):                    
                rgb_series[b] += [y]

    # vérification de la présence des bandes
    for s in series:
        liste = series[s]
        if len(liste) == 0:
            print("[FATAL] Missing band %s" % (s))
            exit()
    print("[INFO] Spectral file loading done.")

    for s in rgb_series:
        liste = rgb_series[s]
        if len(liste) == 0:
            print("[FATAL] Missing band %s" % (s))
            exit()
    print("[INFO] RGB file loading done.")
                
    return series, gt_series, rgb_series


def float_range(mini, maxi):
    """
    Get a choice of range.
    """
    def float_range_checker(arg):
        try:
            f = float(arg)
        except ValueError:
            raise argparse.ArgumentTypeError("must be a floating value")
        if f < mini or maxi < f:
            raise argparse.ArgumentTypeError("must be in range [" + str(mini)
                                             + ", " + str(maxi) + "].")
        return f
    return float_range_checker


def demo():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="action")

    help_detect="Rien."

    bandes = ['B01', 'B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'B08', 'B8A',
              'B09', 'B10', 'B11', 'B12']

    a = subparsers.add_parser("mire", help="Crée une série de mires.")
    a.add_argument("--n", type=int, required=False, default=3,
                   help="Nombre de dates.")

    c = subparsers.add_parser("growing",
                              help="partionnement par croissance de région.")
    c.add_argument("--zip", type=str, required=True,
                   help="Archive contenant les fichiers.")
    c.add_argument("--quantile", type=float_range(0., 1.), required=True, 
                   help="Quantile associé au test du χ2.")    
    c.add_argument("--outdir", type=str, required=True, help="…")
    c.add_argument("--stand", type=str, required=True, choices=['G', 'D', 'N'],
                   help="Standardization : global, date by date, none")
    for b in bandes:
        c.add_argument('--' + b, type=str, choices=['True', 'False'], help="…",
                   required=False, default='False')

    d = subparsers.add_parser("growingit",
                              help="partionnement par croissance de région.")
    d.add_argument("--zip", type=str, required=True,
                   help="Archive contenant les fichiers.")
    d.add_argument("--quantile", type=float_range(0., 1.), required=True, 
                   help="Quantile associé au test du χ2.")    
    d.add_argument("--outdir", type=str, required=True, help="…")
    d.add_argument("--stand", type=str, required=True, choices=['G', 'D', 'N'],
                   help="Standardization : global or date by date, none")
    for b in bandes:
        d.add_argument('--' + b, type=str, choices=['True', 'False'], help="…",
                   required=False, default='False')

    e = subparsers.add_parser("unmixing",
                              help="Démixage de sources.")
    e.add_argument("--zip", type=str, required=True,
                   help="Archive contenant les fichiers.")
    e.add_argument("--outdir", type=str, required=True, help="…")
    e.add_argument("--stand", type=str, required=True, choices=['G', 'D', 'N'],
                   help="Standardization : global or date by date, none")
    for b in bandes:
        e.add_argument('--' + b, type=str, choices=['True', 'False'], help="…",
                   required=False, default='False')

    f = subparsers.add_parser("semi", help="Démixage de source, spectre fixé.")
    f.add_argument("--zip", type=str, required=True,
                   help="Archive contenant les fichiers.")
    f.add_argument("--outdir", type=str, required=True, help="…")
    f.add_argument("--stand", type=str, required=True, choices=['G', 'D', 'N'],
                   help="Standardization : global or date by date, none")
    f.add_argument("--n0", type=int, required=True,
                   help="Instant de calcul du spectre.")
    f.add_argument("--side", type=int, required=True,
                   help="Taille de la vignette.")                  
                  
    for b in bandes:
        f.add_argument('--' + b, type=str, choices=['True', 'False'], help="…",
                   required=False, default='False')
                  
    z = subparsers.add_parser("demo", help=help_detect)
    z.add_argument("--zip", type=str, required=True,
                   help="Archive contenant les fichiers.")
    z.add_argument("--method", type=str, required=True, choices=['G', 'D'],
                   help="Clustering method : global or date by date")
    z.add_argument("--stand", type=str, required=True, choices=['G', 'D'],
                   help="Standardization : global or date by date")
    z.add_argument("--Q", type=float_range(0.1, 0.5), required=True, 
                   help="Proportion of pixels to use to estimate the GMM.")    
    z.add_argument("--clusters", type=int, required=True,
                   help="Cluster number.")
    z.add_argument("--outdir", type=str, required=True, help="…")
    z.add_argument("--quantile", type=float, help="…", required=False, default=-1.0)   
    z.add_argument("--dyna", type=str, choices=['I', 'G'], required=False,
                   help="Type d'étirement de la dynamique.", default='G')    
    for b in bandes:
        z.add_argument('--' + b, type=str, choices=['True', 'False'], help="…",
                   required=False, default='False')
        
    cmd = parser.parse_args()

    if cmd.action == "semi":
        np.set_printoptions(precision=4)
        # chargement des fichiers
        series, gt_series, rgb = load_series(cmd)
        
        algo_info = open("algo_info.txt","w")
        algo_info.write('nim='+str(len(rgb['B02']))+'\n')
        algo_info.close()

        # mise à l'échelle par interpolation bicubique
        series, ndim = interpoler(series, join(cmd.outdir, "interpolated"))

        # démixage
        print("démixer")
        semiunmixing(join(cmd.outdir, "semi"), series, gt_series, cmd.n0,
                     cmd.stand, cmd.side)
                  
        # enregistrement des fichiers de sortie
        vide = np.array([])
        enregistrer([], rgb, None, 0, cmd.outdir, [], [], [], [], [], [])
        exit()

    if cmd.action == "unmixing":
        np.set_printoptions(precision=4)
        
        # chargement des fichiers
        series, gt_series, rgb = load_series(cmd)
        
        algo_info = open("algo_info.txt","w")
        algo_info.write('nim='+str(len(rgb['B02']))+'\n')
        algo_info.close()

        # mise à l'échelle par interpolation bicubique
        series, ndim = interpoler(series, join(cmd.outdir, "interpolated"))

        mappe = calculer_zones_interdites(series, -1)
        # démixage
        print("démixer")
        unmixing(join(cmd.outdir, "unmixing"), mappe, series, gt_series, cmd.stand, ndim)
        
        # enregistrement des fichiers de sortie
        vide = np.array([])
        enregistrer([], rgb, None, 0, cmd.outdir, [], [], [], [], [], [])
        exit()
        
    if cmd.action == "growingit":
        np.set_printoptions(precision=4)
        
        # chargement des fichiers
        series, gt_series, rgb = load_series(cmd)
        
        algo_info = open("algo_info.txt","w")
        algo_info.write('nim='+str(len(rgb['B02']))+'\n')
        algo_info.close()

        # mise à l'échelle par interpolation bicubique
        series = interpoler(series, join(cmd.outdir, "interpolated"))

        # calcul du masque des zones interdites
        mappe = calculer_zones_interdites(series, -1)
        iio.write("mappe.png", np.array(mappe, dtype=int))

        # partionnement par croissance de région
        msq_series, gt_series = growing_iteratively(join(cmd.outdir, "growing"),
                                                    series, mappe, gt_series,
                                                    cmd.stand, cmd.quantile)
        # performances
        performances(cmd.outdir, msq_series, gt_series, cmd.quantile)

        # enregistrement des fichiers de sortie
        vide = np.array([])
        enregistrer([], rgb, None, 0, cmd.outdir, [], [], [], [], [], [])
        
        exit()

    if cmd.action == "growing":
        np.set_printoptions(precision=4)
        
        # chargement des fichiers
        series, gt_series, rgb = load_series(cmd)
        
        algo_info = open("algo_info.txt","w")
        algo_info.write('nim='+str(len(rgb['B02']))+'\n')
        algo_info.close()

        # mise à l'échelle par interpolation bicubique
        series = interpoler(series, join(cmd.outdir, "interpolated"))

        # calcul du masque des zones interdites
        mappe = calculer_zones_interdites(series, -1)
        iio.write("mappe.png", np.array(mappe, dtype=int))

        # partionnement par croissance de région
        msq_series, gt_series = growing(join(cmd.outdir, "growing"),
                                        series, mappe, gt_series,
                                        cmd.stand, cmd.quantile)
        # performances
        performances(cmd.outdir, msq_series, gt_series, cmd.quantile)

        # enregistrement des fichiers de sortie
        enregistrer([], rgb, None, 0, cmd.outdir, [], [], [], [], [], [])

        exit()
        
    if cmd.action == "mire":
        mire(cmd.n)
        exit()
        
    if cmd.action == "demo":
        np.set_printoptions(precision=4)
        
        # chargement des fichiers
        series, gt_series, rgb = load_series(cmd)
        
        algo_info = open("algo_info.txt","w")
        algo_info.write('nim='+str(len(rgb['B02']))+'\n')
        algo_info.close()

        # mise à l'échelle par interpolation bicubique
        series = interpoler(series, join(cmd.outdir, "interpolated"))

        # calcul du masque des zones interdites
        mappe = calculer_zones_interdites(series, cmd.quantile)
        iio.write("mappe.png", np.array(mappe, dtype=int))
#        exit()
        # classification par mélange de gaussiennes            
        res = classifier_espace(join(cmd.outdir, "classified"), series, mappe,
                                cmd.clusters, cmd.method, cmd.stand, cmd.Q)
            
        [neofiles, bandes, spectres_μ, spectres_σ, spectres_cσ, coords_μ] = res
        mat_perm = None
#COM        # calculer les spectres de chaque grappe
#COM        if cmd.desc == 'μ' or cmd.desc == 'mu':
#COM            descripteur = 'M'
#COM        if cmd.desc == 'σ' or cmd.desc == 'sig':
#COM            descripteur = 'S'
#COM        if cmd.desc == 'μσ' or cmd.desc == 'musig':
#COM            descripteur = 'MS'
        
        palette = choisir_palette(len(neofiles), cmd.clusters, coords_μ, cmd.dyna)
        # enregistrement des fichiers de sortie
        enregistrer(neofiles, rgb, gt_series, cmd.clusters, cmd.outdir, spectres_μ,
                    spectres_σ, spectres_cσ, bandes, palette, mat_perm)
    plt.show()
    return


def mire(ndates):

    zoom = {'B01': 6, 'B02': 1, 'B03': 1, 'B04': 1, 'B05': 2, 'B06': 2,
            'B07': 2, 'B08': 1, 'B8A': 2, 'B09': 6, 'B10': 6, 'B11': 2,
            'B12': 2}
    
    val = [30, 50, 80, 120, 170, 230]
#    σ = [1, 3, 6, 10, 15, 21]
    σ = [1, 1, 1, 1, 1, 1]

    for t in np.arange(ndates):
        for z in zoom:
            nlig, ncol = 240 // zoom[z], 360 // zoom[z]
            cote = 120 // zoom[z]

            # image à l'échelle 1
            im = np.zeros((nlig, ncol))

            nc = np.array(np.array([0, 120, 240, 0, 120, 240])/zoom[z], dtype=int)
            nl = np.array(np.array([0, 0, 0, 120, 120, 120])/zoom[z], dtype=int)

            for i in np.arange(6):
                im[nl[i]:nl[i]+cote, nc[i]:nc[i]+cote] = \
                val[i] + σ[i] * np.random.randn(cote, cote)

            iio.write("mire_%02d_%s.tif" %(t, z), im)
    return

def verifier():

    B=3
    M=5
    λ = np.random.rand(1, 1)
    p = np.random.rand(B, 1)
    P = np.kron(p, np.ones(M))
    
    X = np.random.rand(B, M)
    W = np.random.rand(B, M)

    v1 = np.trace((λ * W + P).T @ (λ * W + P))

    U = np.concatenate((λ, p), axis=0)

    β = np.trace(W.T@W)

    α = np.sum(W, axis=1)

    A = M * np.identity(B+1)
    A[0, 0] = β
    A[0, 1:] = α
    A[1:, 0] = α

    v2 = U.T@A@U
    print(λ)
    print(p)
    print(X)
    print(W)
    print(U)
    print(A)

    print(v1, v2)
    return


def estimate_λ(xn0, xn):
    """
    Parameters
    ----------
    xn0 : np.array, ndim=(nlig, ncol, nbandes)
        Matrix representing the feature sample at date n0.

    xn : np.array, ndim=(nlig, ncol, nbandes)
        Matrix representing the feature sample at date n.

    Returns
    -------
    λ : float
        Parameter λ.
    """

    nlig, ncol, nbandes = xn0.shape
    x = np.reshape(xn0, (nlig*ncol, nbandes))
    y = np.reshape(xn, (nlig*ncol, nbandes))
    λ = np.trace(x @ y.T) / np.trace(y @ y.T)
    
    return λ


def compute_λ_map_series(X, GT, spectrum_p, side, nlig, ncol, outdir):
    """
    Parameters
    ----------
    X : np.array, ndim=(ndates, nlig*ncol, nbandes)
        Matrix representing the feature sample at different dates.
    p : np.array, ndim=(nbandes)
        Solar panel spectrum.
    side : int
        Side of the square patch.
    nlig : int
        Number of lines of the images.
    ncol : int
        Number of columns of the images.
    outdir : str
        Output directory.

    Returns
    -------
    λ_map_files : list
        List of the filename of the λ maps.
    """

    ndates, npix, nbandes = X.shape
    S = np.reshape(X, (ndates, nlig, ncol, nbandes))
    GT = np.reshape(GT, (ndates, nlig, ncol))
    
    λ_map_files = []
    for n0 in np.arange(ndates):
        print("date %d" % (n0))
        λ_map = np.zeros((nlig, ncol))
#        if n0 == 0:
#            name = join(outdir, "mappe_%d.tif" % (n0))
#            iio.write(name, λ_map)
#            λ_map_files += [name]
#            continue
        
        for l in np.arange(0, nlig, side):
            for c in np.arange(0, ncol, side):
                lmax = nlig if l + side > nlig else l + side
                cmax = ncol if c + side > ncol else c + side
                x0 = S[n0, l:lmax, c:cmax, :] - spectrum_p

                λ = []
                if n0 == 0:
                    xn = S[n0, l:lmax, c:cmax, :]                    
                    λ += [estimate_λ(x0, xn)]
                else:
                    for n in np.arange(n0):
                        xn = S[n, l:lmax, c:cmax, :]                    
                        λ += [estimate_λ(x0, xn)]
                λ = np.median(λ)
                λ_map[l:l+side, c:c+side] = λ
        name = join(outdir, "mappe_%d.tif" % (n0))
        iio.write(name, λ_map)
        panel_λ = λ_map[GT[n0] == True]
        ground_λ = λ_map[GT[n0] == False]
        plt.figure(1)
        plt.clf()
        p_λ, bins = np.histogram(panel_λ, bins=100, range=[-1, 2])
        g_λ, bins = np.histogram(ground_λ, bins=100, range=[-1, 2])
        plt.title("Repartition of the λ values.")
        plt.plot(bins[:-1], p_λ, label="λ values on panels")
        plt.plot(bins[:-1], g_λ, label="λ values on ground")
#        plt.hist(panel_λ, bins=100, range=[-1, 2], label="λ values on panels")
#        plt.hist(ground_λ, bins=100, range=[-1, 2], label="λ values on ground")
        plt.grid(True)
        plt.legend()
        plt.savefig(join(outdir, "histolamb_%d.png" % (n0)))
        
        λ_map_files += [name]
    return λ_map_files


def rien():

    plt.figure(1)
    graphe = plt.subplot(2, 1, 1)
    graphe.grid(True)
    graphe.set_ylim([0, 10])
    graphe.set_xticks(np.arange(1, 4), labels=['a', 'b', 'c'])
 
    graphe.plot([1, 2, 3], [2, 4, 6], label="rien")
    
    graphe.legend()

    graphe = plt.subplot(2, 1, 2)
    graphe.plot([1, 2, 3], [2, 4, 6])

    plt.figure(2)
    plt.plot([1, 2, 3], [2, 4, 6])
    print(graphe)

    return
    
    
if __name__ == "__main__":
#    mire();  exit()
# ALT    for i in np.arange(3):
# ALT        rien()
# ALT    plt.show()
# ALT    exit()
    np.random.seed(0)
#    verifier(); exit()
    print("#######################")
    print("##### VERSION 0.6 #####")
    print("#######################")

    demo()
#    pied()
#    main()
    exit()

"""    
2021-06-15
2022-01-31
ls *D01* | tail -n +68 | head -22

Résultats assez corrects avec :
--B04 True --B03 True --B02 True --gmm S --clusters 6

python3 $pgm demo --zip con3.zip --gmm S --outdir testeur --B01 True --B02 True --B03 True --B04 True --B05 True --B06 True --B07 True --B08 True --B8A True --B09 True --B10 True --B11 True --B12 True --clusters 6;

python3 ../developpement/solarium-0.5/solarium.py demo --zip ~/ovh/projets/solarium/images/ipol/bluebell_2.zip --outdir ici0.5 --B01 True --B02 True --B03 True --B04 True --B05 True --B06 True --B07 True --B08 True --B8A True --B09 True --B10 True --B11 True --B12 True --clusters 8 --dyna G --method "D" --stand "D" --Q 0.25

~/ENS/projets/solarium/experiences$ source EnvSol/bin/activate; unset PYTHONPATH;

python ../developpement/solarium-0.7/solarium.py unmixing --zip holstein_1_solar_farm.zip --stand D --outdir bof --B01 True --B02 True --B03 True --B04 True --B05 True --B06 True --B07 True --B08 True --B8A True --B09 True --B10 True --B11 True --B12 True;
"""

"""
Cher Jean-Michel,

je te fait part de mes réflexions car malgré la simplicité de l'énoncé
du problème de colorisation 1, celui-ci ne m'apparaît pas si simple
que cela.

En effet, ta suggestion de résolution revient à trouver une
application injective dans la mesure où l'on pourrait avoir deux
clusters de t+1 ayant le même plus proche cluster de t. Ceci aurait
donc pour conséquence de diminuer éventuellement le nombre de couleurs
dans l'image t+1. S'autorise-t-on cela ou bien souhaite-t-on conserver
le même nombre de couleurs à chaque date ?

Il me semble que l'on veut conserver le même nombre de couleurs. Par
conséquent, on cherche une relation bijective. J'ai d'abord pensé à
une méthode naïve en O(N^3) où N est le nombre de clusters :

(solution A)
0) calculer la matrice des distances M_{N, N}
1) chercher le minimum absolu dans M_{N, N}
2) associer le couple C_i(t+1), C_j(t) correspondant 
3) réduire la matrice des distances à M_{N-1, N-1} en interdisant que 
les nouvelles associations n'utilisent les C_i(t+1), C_j(t) précédents 
4) itérer en revenant à 1)

Mais il m'apparaît qu'il s'agit en fait d'un problème de permutation
optimal, dont la complexité est en O(N^N). Si l'on reste sur un nombre
raisonnable de clusters, i.e. N<=7, on peut employer une recherche par
force brute (solution B). En revanche, si l'on passe au-dessus, il faut 
chercher une méthode d'approximation (solution C) plus complexe à mettre
en œuvre.

Si tu souhaites une réalisation rapide, je peux te fournir A ou B
(mais plus long en temps de calcul) cette semaine. Mais si tu
souhaites une méthode plus générale, il faut que je regarde la
littérature.

"""
"""
Hydrogene1?
"""

